#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*
 *  long_spanning_reads.cpp
 *  TopHat
 *
 *  Created by Cole Trapnell on 2/5/09.
 *  Copyright 2009 Cole Trapnell. All rights reserved.
 *
 */

#include <cassert>
#include <cstdio>
#include <vector>
#include <string>
#include <map>
#include <algorithm>
#include <set>
//#include <stdexcept>
#include <iostream>
#include <fstream>
#include <cstring>
#include <seqan/sequence.h>
#include <seqan/find.h>
#include <seqan/file.h>
//#include <seqan/modifier.h>
#include <getopt.h>

#include "common.h"
#include "bwt_map.h"
#include "tokenize.h"
#include "segments.h"
#include "reads.h"
#include "junctions.h"

using namespace seqan;
using namespace std;

void print_usage()
{
    fprintf(stderr, "Usage:   segment_juncs <ref.fa> <segment.juncs> <left_seg1.bwtout,...,segN.bwtout> [right_seg1.bwtout,...,right_segN.bwtout]\n");
}

const char *short_options = "i:I:fq";

// This is the maximum number of bowtie mismatches allower per segment hit
static const int num_bowtie_mismatches = 2;
uint32_t max_cov_intron_length = 3000;
uint32_t max_seg_intron_length = 300000;
static const int max_cov_juncs = 5000000;
static const int max_seg_juncs = 10000000;

static struct option long_options[] = {
{"min-anchor",       required_argument,       0,            'a'},
{"min-intron",       required_argument,       0,            'i'},
{0, 0, 0, 0} // terminator
};

int parse_options(int argc, char** argv)
{
    int option_index = 0;
    int next_option;
    do {
        next_option = getopt_long(argc, argv, short_options, long_options, &option_index);
        switch (next_option) {
			case 'i':
				min_intron_length = (uint32_t)parseInt(1, "-a/--min-intron arg must be at least 1", print_usage);
				break;
			case 'C':
				max_cov_intron_length = parseInt(1,"-I arg must be at least 1", print_usage);
				break;
			case 'I':
				max_intron_length = parseInt(1,"-I arg must be at least 1", print_usage);
				break;
			case 'f':
				reads_format = FASTA;
				break;
			case 'q':
				reads_format = FASTQ;
				break;
			case -1:     /* Done with options. */
                break;
            default:
                print_usage();
                return 1;
        }
    } while(next_option != -1);
    
    return 0;
}

void get_seqs(istream& ref_stream,
			  RefSequenceTable& rt,
			  bool keep_seqs = true,
			  bool strip_slash = false)
{    
    while(ref_stream.good() &&
          !ref_stream.eof())
    {
		RefSequenceTable::Sequence* ref_str = new RefSequenceTable::Sequence();
        string name;
        readMeta(ref_stream, name, Fasta());
		fprintf(stderr, "\tLoading %s...", name.c_str());
		seqan::read(ref_stream, *ref_str, Fasta());
		fprintf(stderr, "done\n");
        rt.get_id(name, keep_seqs ? ref_str : NULL);
		if (!keep_seqs)
			delete ref_str;
    }	
}


RefSeg seg_from_bowtie_hit(const BowtieHit& T)
{
	RefSeg r_seg(T.ref_id(), false, 0,0);
	
	if (T.antisense_align())
	{
		r_seg.left = max(0, T.right() - 2);
		r_seg.right = T.right() + (T.right() - T.left() + 1); // num allowed bowtie mismatches 
		r_seg.points_left = false;
	}
	else
	{
		r_seg.left = max(0, T.left() - (T.right() - T.left() + 1));
		r_seg.right = T.left() + 2; // num allowed bowtie mismatches
		r_seg.points_left = true;
	}
	
	return r_seg;
}

pair<RefSeg, RefSeg> segs_from_bowtie_hits(const BowtieHit& T,
										   const BowtieHit& H)
{
	pair<RefSeg, RefSeg> seg_pair;
	if (H.antisense_align() == false && 
		abs((H.right() + 1) - T.left()) < (int)max_intron_length)
	{
		RefSeg left_seg(H.ref_id(), false, 0,0);
		left_seg.left = max(0, H.right() - 2);
		left_seg.right = H.right() + (H.right() - H.left() + 1); // num allowed bowtie mismatches 
		
		RefSeg right_seg(T.ref_id(), true, 0,0);
		right_seg.left = max(0, T.left() - (T.right() - T.left() + 1));
		right_seg.right = T.left() + 2; // num allowed bowtie mismatches
		
		seg_pair = make_pair(left_seg, right_seg);
	}
	else if (H.antisense_align() == true &&
			 abs((T.right() + 1) - H.left()) < (int)max_intron_length)
	{
		RefSeg left_seg(T.ref_id(), false, 0,0);
		left_seg.left = max(0, T.right() - 2);
		left_seg.right = T.right() + (T.right() - T.left() + 1); // num allowed bowtie mismatches 
		
		RefSeg right_seg(H.ref_id(), true, 0,0);
		right_seg.left = max(0, H.left() - (H.right() - H.left() + 1));
		right_seg.right = H.left() + 2; // num allowed bowtie mismatches
		
		seg_pair = make_pair(left_seg, right_seg);
	}
	return seg_pair;
}


struct IntronMotifs
{
	IntronMotifs(uint32_t rid) : ref_id(rid) {}
	uint32_t ref_id;
	vector<size_t> fwd_donors;
	vector<size_t> fwd_acceptors;
	vector<size_t> rev_donors;
	vector<size_t> rev_acceptors;	
	
	void finalize(vector<size_t>& f)
	{
		sort(f.begin(), f.end());
		vector<size_t>::iterator i = unique(f.begin(), f.end());
		f.erase(i, f.end());
	}
	
	void finalize()
	{
		finalize(fwd_donors);
		finalize(fwd_acceptors);
		finalize(rev_donors);
		finalize(rev_acceptors);
	}
};

template <class JunctionSet>
void juncs_from_ref_segs(RefSequenceTable& rt,
						 vector<RefSeg>& expected_don_acc_windows,
						 JunctionSet& juncs,
						 bool suppress_long = false,
						 int max_intron = max_intron_length,
						 size_t max_juncs = 0xFFFFFFFF)
{	

	typedef map<uint32_t, IntronMotifs> MotifMap;
	
	MotifMap ims;
	
	//fprintf(stderr, "Collecting motifs...\n");
	
	for (size_t r = 0; r < expected_don_acc_windows.size(); ++r)
	{
		const RefSeg& seg = expected_don_acc_windows[r];
		
		RefSequenceTable::Sequence* ref_str = rt.get_seq(seg.ref_id);
		
		if (!ref_str)
		{
			fprintf(stderr, "Error: reference sequence for ID %d\n, not found, aborting\n", seg.ref_id);
			exit(1);
		}
		
		pair<map<uint32_t, IntronMotifs>::iterator, bool> ret = 
			ims.insert(make_pair(seg.ref_id, IntronMotifs(seg.ref_id)));
		
		IntronMotifs& motifs = ret.first->second;
		
		RefSequenceTable::Sequence seg_str = seqan::infix(*ref_str, 
														  seg.left, 
														  seg.right);
		
		assert (seg.left >= 0 && seg.right < (int)length(*ref_str) - 1);
		//		if (seg.left > 100129650 && seg.left < 100129725 && seg.right > 100133930 &&  seg.right < 100133990)
		//cout << seg.left << " " << seg.right << " " << seg_str << endl;
		if (seg.points_left)
		{
			// A ref segment that "points left" is one that was flanked 
			// on the right by a partial bowtie hit, indicating that we
			// should be looking for an intron to the left of the hit
			// In this seg, that means either an "AG" or an "AC"
			
			
			for (size_t i = 0; i < length(seg_str) - 1; ++i)
			{
				// Look at a slice of the reference without creating a copy.
				RefSequenceTable::Sequence curr = seqan::infix(seg_str, 
															   i, 
															   i + 2);
				if (curr == "AG")
					motifs.fwd_acceptors.push_back(seg.left + i);
				else if (curr == "AC")
					motifs.rev_donors.push_back(seg.left + i);
			}
		}
		else
		{
			// A right pointing ref seg wants either a "GT" or a "CT"
			for (size_t i = 0; i < length(seg_str) - 1; ++i)
			{
				// Look at a slice of the reference without creating a copy.
				RefSequenceTable::Sequence curr = seqan::infix(seg_str, 
															   i, 
															   i + 2);
				if (curr == "GT")
					motifs.fwd_donors.push_back(seg.left + i);
				else if (curr == "CT")
					motifs.rev_acceptors.push_back(seg.left + i);
			}
		}
	}
	
	int num_juncs = 0;
	
	//fprintf(stderr, "reporting synthetic splice junctions...\n");
	for (MotifMap::iterator motif_itr = ims.begin(); motif_itr != ims.end(); ++motif_itr)
	{
		uint32_t ref_id = motif_itr->first;
		IntronMotifs& motifs = motif_itr->second;
		motifs.finalize();
		
		vector<size_t>& fwd_donors = motifs.fwd_donors;
		vector<size_t>& fwd_acceptors = motifs.fwd_acceptors;
		vector<size_t>& rev_acceptors = motifs.rev_acceptors;
		vector<size_t>& rev_donors = motifs.rev_donors;

	
		//const char* ref_name = rt.get_name(motif_itr->second.ref_id);
		
		for (size_t d = 0; d < fwd_donors.size(); ++d)
		{
			// start pos is a lower bound on downstream acceptor positions
			// to consider
			size_t start_pos = fwd_donors[d] + min_intron_length;
			
			vector<size_t>::iterator lb = upper_bound(fwd_acceptors.begin(), 
													  fwd_acceptors.end(),
													  start_pos);
			
			if (lb == fwd_acceptors.end())
				break;
			for (size_t a = lb - fwd_acceptors.begin(); 
				 a < fwd_acceptors.size(); 
				 ++a)
			{
				
				if (suppress_long && 
					(int)fwd_acceptors[a] - (int)fwd_donors[d] > max_intron)
				{
					break;
				}
				num_juncs++;
				juncs.insert(Junction(ref_id,
									  fwd_donors[d] - 1,
									  fwd_acceptors[a] + 2,
									  false,
									  a));

				if (juncs.size() > max_juncs)
				{
					juncs.erase(*(juncs.rbegin()));
				}			
			}
		}
		
		for (size_t a = 0; a < rev_acceptors.size(); ++a)
		{
			
			// start pos is a lower bound on downstream acceptor positions
			// to consider
			size_t start_pos = rev_acceptors[a] + min_intron_length;
			
			vector<size_t>::iterator lb = upper_bound(rev_donors.begin(), 
													  rev_donors.end(),
													  start_pos);
			
			if (lb == rev_donors.end())
				break;
			for (size_t d = lb - rev_donors.begin(); 
				 d < rev_donors.size(); 
				 ++d)
			{
				
				if (suppress_long && 
					(int)rev_donors[d] - (int)rev_acceptors[a] > max_intron)
				{
					break;
				}
				num_juncs++;
				juncs.insert(Junction(ref_id,
									  rev_acceptors[a] - 1,
									  rev_donors[d] + 2,
									  true,
									  d));
				if (juncs.size() > max_juncs)
				{
					juncs.erase(*(juncs.rbegin()));
				}
			}
		}
	}
	//fprintf(stderr, "Found %d total splices\n", num_juncs);
}

struct skip_count_lt
{
	bool operator()(const Junction& lhs, const Junction& rhs)
	{
		if (lhs.skip_count != rhs.skip_count)
			return lhs.skip_count < rhs.skip_count;
		return lhs < rhs;
	}
};

void find_gaps(RefSequenceTable& rt,
			   vector<HitsForRead>& hits_for_read,
			   std::set<Junction, skip_count_lt>& seg_juncs)
{
	vector<RefSeg> expected_don_acc_windows;
	
	if (hits_for_read.empty())
		return;
	size_t last_non_empty = hits_for_read.size() - 1;
	while(last_non_empty >= 0 && hits_for_read[last_non_empty].hits.empty())
	{
		--last_non_empty;
	}
	
	hits_for_read.resize(last_non_empty + 1);
	for (size_t s = 0; s < hits_for_read.size(); ++s)
	{
		HitsForRead& curr = hits_for_read[s];
		
		for (size_t h = 0; h < curr.hits.size(); ++h)
		{
			bool found_left_partner = s == 0;
			bool found_right_partner = s == hits_for_read.size() - 1;
			bool found_distant_partner = false;
			BowtieHit& bh = curr.hits[h];
			if (s > 0)
			{
				// Look for a left partner for the current hit
				HitsForRead& left = hits_for_read[s - 1];
				
				for (size_t l = 0; l < left.hits.size(); ++l)
				{
					BowtieHit& lh = left.hits[l];
					if (bh.antisense_align() != lh.antisense_align())
						continue;
					if ((bh.antisense_align() && bh.right() == lh.left()) ||
						(!bh.antisense_align() && lh.right() == bh.left() ))
					{
						found_left_partner = true;
						break;
					}
					if ((bh.antisense_align() && abs(bh.right() - lh.left()) < (int)max_intron_length) ||
						(!bh.antisense_align() && abs(lh.right() - bh.left()) < (int)max_intron_length))
					{
						found_distant_partner = true;
					}
						
				}
			}
			if (s < hits_for_read.size() - 1)
			{
				// Look for a right partner for the current hit
				HitsForRead& right = hits_for_read[s + 1];
				
				for (size_t r = 0; r < right.hits.size(); ++r)
				{
					BowtieHit& rh = right.hits[r];
					if (bh.antisense_align() != rh.antisense_align())
						continue;
					if ((bh.antisense_align() && rh.right() == bh.left()) ||
						(!bh.antisense_align() && bh.right() == rh.left() ))
					{
						found_right_partner = true;
						break;
					}
					if ((bh.antisense_align() && abs(rh.right() - bh.left()) < (int)max_intron_length) ||
						(!bh.antisense_align() && abs(bh.right() - rh.left()) < (int)max_intron_length))
					{
						found_distant_partner = true;
					}
				}
			}
			if (!found_left_partner)
			{
				if (!bh.antisense_align())
				{
					RefSeg left_seg(bh.ref_id(), true, 0,0);
					//if (found_distant_partner)
					//	left_seg.left = max(0, bh.left() - 3);
					//else
					left_seg.left = max(0, bh.left() - (bh.read_len() + 1));
					left_seg.right = bh.left() + 5; // num allowed bowtie mismatches
					expected_don_acc_windows.push_back(left_seg);
				}
				else
				{
					RefSeg right_seg(bh.ref_id(), false, 0,0);
					right_seg.left = max(0, bh.right() - 5);
					//if (found_distant_partner)
					//	right_seg.right = bh.right() + 3;
					//else
					right_seg.right = bh.right() + (bh.read_len() + 1); // num allowed bowtie mismatches
					expected_don_acc_windows.push_back(right_seg);
				}
			}
			if (!found_right_partner)
			{
				if (!bh.antisense_align())
				{
					RefSeg right_seg(bh.ref_id(), false, 0,0);
					right_seg.left = max(0, bh.right() - 5);
					//if (found_distant_partner)
					//	right_seg.right = bh.right() + 3;
					//else
					right_seg.right = bh.right() + (bh.read_len() + 1); // num allowed bowtie mismatches
					expected_don_acc_windows.push_back(right_seg);
				}
				else
				{
					RefSeg left_seg(bh.ref_id(), true, 0,0);
					//if (found_distant_partner)
					//	left_seg.left = max(0, bh.left() - 3);
					//else
					left_seg.left = max(0, bh.left() - (bh.read_len() + 1));
					left_seg.right = bh.left() + 5; // num allowed bowtie mismatches
					expected_don_acc_windows.push_back(left_seg);	
				}
			}
			
		}
		
	}
	
	juncs_from_ref_segs(rt, expected_don_acc_windows, seg_juncs, true, max_seg_intron_length, max_seg_juncs);

	//juncs_from_ref_segs(rt, expected_don_acc_windows, juncs);
}

void look_for_hit_group(RefSequenceTable& rt,
						ReadTable& unmapped_reads,
						vector<HitStream>& seg_files,
						int curr_file,
						uint64_t insert_id,
						vector<HitsForRead>& hits_for_read,
						std::set<Junction, skip_count_lt>& juncs)
{
	HitStream& curr = seg_files[curr_file];
	
	HitsForRead hit_group;
	
	int order = unmapped_reads.observation_order(insert_id);
	
	while(true)
	{
		uint64_t next_group_id = curr.next_group_id();
		int next_order = unmapped_reads.observation_order(next_group_id);
		
		// If we would have seen the hits by now, stop looking in this stream,
		// but forward the search if possible.
		if (order < next_order || next_group_id == 0)
		{		
			if (curr_file > 0)
			{
				look_for_hit_group(rt,
								   unmapped_reads,
								   seg_files,
								   curr_file - 1,
								   insert_id,
								   hits_for_read,
								   juncs);
			}
			break;
		}
		else if (curr.next_read_hits(hit_group))
		{

			// If we found hits for the target group in the left stream,
			// add them to the accumulating vector and continue the search
			if (hit_group.insert_id == insert_id)
			{	
				hits_for_read[curr_file] = hit_group;
				if (curr_file > 0)
				{
					// we need to look left (recursively) for the group we 
					// just read for this stream.
					look_for_hit_group(rt,
									   unmapped_reads,
									   seg_files,
									   curr_file - 1,
									   insert_id,
									   hits_for_read,
									   juncs);
					
				}
				break;
			}
			else if (curr_file > 0)
			{
				// If this is a different group, we need to start a whole new 
				// search for it, with a whole new vector of hits.
				vector<HitsForRead> hits_for_new_read(seg_files.size());
				hits_for_new_read[curr_file] = hit_group;

				
				look_for_hit_group(rt,
								   unmapped_reads,
								   seg_files,
								   curr_file - 1,
								   hit_group.insert_id,
								   hits_for_new_read,
								   juncs);
				find_gaps(rt, hits_for_new_read, juncs);
			}

		}
	}
	
	
}


bool process_next_hit_group(RefSequenceTable& rt,
							ReadTable& unmapped_reads,
							vector<HitStream>& seg_files, 
							size_t curr_file,
							std::set<Junction, skip_count_lt>& juncs)
{
	HitStream& curr = seg_files[curr_file];
	HitsForRead hit_group;
	if (curr.next_read_hits(hit_group))
	{
		vector<HitsForRead> hits_for_read(seg_files.size());
		hits_for_read.back() = hit_group;

		//int ord = unmapped_reads.observation_order(hit_group.insert_id);
		look_for_hit_group(rt,
						   unmapped_reads, 
						   seg_files, 
						   (int)curr_file - 1,
						   hit_group.insert_id,
						   hits_for_read,
						   juncs);
		find_gaps(rt, hits_for_read, juncs);
		return true;
	}
	return false;
}

static const int UNCOVERED = 0;
static const int LOOK_LEFT = 1;
static const int LOOK_RIGHT = 2;

uint8_t get_cov(const vector<uint8_t>& cov, uint32_t c)
{
	uint32_t b = c >> 2;
	uint32_t r = c & 0x3;
	uint8_t  s = (r << 1);
	uint8_t  v = cov[b];
	v &= (0x3 << s);
	v >>= s;
	return v;
}

void capture_island_ends(ReadTable& it,
						 RefSequenceTable& rt,
						 vector<FILE*>& seg_files,
						 std::set<Junction, skip_count_lt>& cov_juncs,
						 int min_intron,
						 int max_intron)
{
	//static int island_repeat_tolerance = 10;
	BowtieHitFactory hit_factory(it,rt);
	map<uint32_t, vector<uint8_t> > coverage_map;
	vector<RefSeg> expected_don_acc_windows;
	
	//FILE* f_islands = fopen("islands.coords", "w");
	
	for (size_t f = 0; f < seg_files.size(); ++f)
	{
		fprintf(stderr, "Adding hits from segment file %d to coverage map\n", (int)f);
		FILE* fp = seg_files[f];
		rewind(fp);
		HitStream hs(fp, &hit_factory, false, false);
		
		HitsForRead hit_group;
		while (hs.next_read_hits(hit_group))
		{
			for (size_t h = 0; h < hit_group.hits.size(); ++h)
			{
				BowtieHit& bh = hit_group.hits[h];
				
				pair<map<uint32_t, vector<uint8_t> >::iterator, bool> ret = 
				coverage_map.insert(make_pair(bh.ref_id(), vector<uint8_t>()));
				vector<uint8_t>& ref_cov = ret.first->second;
				
				size_t right_extent = bh.right() >> 2;
				
				if (right_extent >= ref_cov.size())
				{
					ref_cov.resize(right_extent + 1, 0);
					
				}
				
				for (uint32_t c = (uint32_t)bh.left(); c < (uint32_t)bh.right(); ++c)
				{
					uint32_t b = c >> 2;
					uint32_t r = c & 0x3;
					uint32_t  s = (r << 1);
					uint32_t  v = ref_cov[b];
					
					v &= ~(0x3 << s);
					
					//int t = ((v & (0x3 << s)) >> s);
					int t = get_cov(ref_cov, c);
					t++;
					uint8_t m = (uint8_t)min(t, 3);
					v |= m << s;
					ref_cov[b] = v;
					//ref_cov[c] = min((int)ref_cov[c] + 1, 0xFF);
				}
			}
		}
	}
	
	static const int min_cov_length = 27;
	
	long covered_bases = 0;
	int long_enough_bases = 0;
	int left_looking = 0;
	int right_looking = 0;
	static const int extend = 10;
	static const int repeat_tol = 5;
	
	int num_islands = 0;
	
	for (map<uint32_t, vector<uint8_t> >::iterator itr = coverage_map.begin();
		 itr != coverage_map.end();
		 ++itr)
	{
		//fprintf (stderr, "Finding pairings in ref seg %u\n", (int)itr->first);
		
		vector<uint8_t>& cov = itr->second;
		vector<bool> long_enough(cov.size() * 4);

		size_t last_uncovered = 0;
		
		static const uint8_t min_cov = 3;
		
		for (size_t c = 1; c < cov.size() * 4; ++c)
		{
			int c_cov = get_cov(cov, c);
			if (c_cov < min_cov || c == (cov.size() * 4) - 1)
			{
				int putative_exon_length = (int)c - (int)last_uncovered;
				int last_pos_cov = get_cov(cov,c - 1);
				if (last_pos_cov >= min_cov && putative_exon_length >= min_cov_length)
				{
					//fprintf (f_islands, "%s\t%d\t%d\n", rt.get_name(itr->first), last_uncovered + 1, c);
					covered_bases += (c + 1 - last_uncovered);
					//fprintf(stderr, "making new segment from %d to %d\n", last_uncovered + 1, c);
					for (int l = (int)c; 
						 l > (int)last_uncovered; 
						 --l)
					{
						long_enough[l] = true;
					}
				}
				last_uncovered = c;
			}
		}
		
		
		size_t last_cov_idx = 0;
		
		vector<bool>& ref_cov = long_enough;
		vector<uint8_t> cov_state(ref_cov.size(), UNCOVERED);
		
		for (size_t c = 1; c < ref_cov.size(); ++c)
		{
			if (ref_cov[c])
			{
				long_enough_bases++;
				if (!ref_cov[c - 1] && last_cov_idx)
				{
					long putative_intron_length = (long)c - last_cov_idx;
					if (putative_intron_length > (int)min_intron && 
						putative_intron_length < (int)max_intron)
					{
						num_islands += 1;
						for (int l = (int)last_cov_idx - repeat_tol; 
							 l >= 0 && l < (int)last_cov_idx + extend && l < (int)cov_state.size(); 
							 ++l)
						{
							cov_state[l] |= LOOK_RIGHT;
						}
						
						for (int r = (int)c - extend; 
							 r >= 0 && r < (int)c + repeat_tol && r < (int)cov_state.size(); 
							 ++r)
						{
							cov_state[r] |= LOOK_LEFT;
						}
					}
				}
				last_cov_idx = c;
			}
		}

		RefSeg* curr_look_left = NULL;
		RefSeg* curr_look_right = NULL;
		for (size_t c = 1; c < cov_state.size(); ++c)
		{
			if (cov_state[c] & LOOK_LEFT)
			{
				left_looking++;
				if (!(cov_state[c-1] & LOOK_LEFT))
				{
					expected_don_acc_windows.push_back(RefSeg(itr->first, true, c, c + 1));
					curr_look_left = &(expected_don_acc_windows.back());
				}
				else
				{
					curr_look_left->right++;
				}
			}
			else
			{
				if ((cov_state[c-1] & LOOK_LEFT))
				{
					curr_look_left = NULL;
				}
			}
			
			if (cov_state[c] & LOOK_RIGHT)
			{
				right_looking++;
				if (!(cov_state[c-1] & LOOK_RIGHT))
				{
					expected_don_acc_windows.push_back(RefSeg(itr->first, false, c, c + 1));
					curr_look_right = &(expected_don_acc_windows.back());
				}
				else
				{
					curr_look_right->right++;
				}
			}
			else
			{
				if ((cov_state[c-1] & LOOK_RIGHT))
				{
					curr_look_right = NULL;
				}
			}
		}
	}
	
	fprintf(stderr, "Map covers %ld bases\n", covered_bases);
	fprintf(stderr, "Map covers %d bases in sufficiently long segments\n", long_enough_bases);
	fprintf(stderr, "Map contains %d good islands\n", num_islands + 1);
	fprintf(stderr, "%d are left looking bases\n", left_looking);
	fprintf(stderr, "%d are right looking bases\n", right_looking);
	
	juncs_from_ref_segs(rt, expected_don_acc_windows, cov_juncs, true, max_cov_intron_length, max_cov_juncs);
	//juncs.insert(island_juncs.begin(), island_juncs.end());
}


void process_segment_hits(RefSequenceTable& rt,
						  vector<FILE*>& seg_files,
						  BowtieHitFactory& hit_factory,
						  ReadTable& it,
						  std::set<Junction, skip_count_lt>& juncs)
{
	
	//get_seqs(reads_stream, it, false, false);
	//get_seqs(reads_file, it, reads_format, false, false);
	
	vector<HitStream> hit_streams;
	for (size_t i = 0; i < seg_files.size(); ++i)
	{
		HitStream hs(seg_files[i], &hit_factory, false, false);
		
		// if the next group id in this stream is zero, it's an empty stream,
		// and we can simply skip it.
		//if (hs.next_group_id() != 0)
		hit_streams.push_back(hs);
	}
	
	
	while (process_next_hit_group(rt,
								  it, 
								  hit_streams, 
								  hit_streams.size() - 1,
								  juncs) == true); 
	
}

void driver(istream& ref_stream,
			FILE* juncs_out,
            vector<FILE*>& left_seg_files,
            vector<FILE*>& right_seg_files)
{
	if (left_seg_files.size() == 0)
	{
		fprintf(stderr, "No hits to process, exiting\n");
		exit(0);
	}
	
	std::set<Junction, skip_count_lt> seg_juncs;
	std::set<Junction, skip_count_lt> cov_juncs;
	
	std::set<Junction> juncs;
	
	RefSequenceTable rt(true, true);

	fprintf (stderr, "Loading reference sequences...\n");
	get_seqs(ref_stream, rt, true, false);
	fprintf(stderr, "done\n");
	
//	SequenceTable it(true, false);

	ReadTable it;

	BowtieHitFactory hit_factory(it,rt);
	
	if (left_seg_files.size() > 1)
	{
		fprintf( stderr, "Loading left segment hits...");
		process_segment_hits(rt,
							 left_seg_files,
							 hit_factory,
							 it,
							 seg_juncs);
		
		fprintf( stderr, "done.\n");
	}
	
	if (right_seg_files.size() > 1)
	{
		fprintf( stderr, "Loading right segment hits...");
		process_segment_hits(rt,
							 right_seg_files,
							 hit_factory,
							 it,
							 seg_juncs);
		fprintf( stderr, "done.\n");
	}
	
	vector<FILE*> all_seg_files;
	copy(left_seg_files.begin(), left_seg_files.end(), back_inserter(all_seg_files));
	copy(right_seg_files.begin(), right_seg_files.end(), back_inserter(all_seg_files));
	
//	// If all we are working with is single-segment reads, we can't lean 
//	// on segment spanning reads, and instead must rely more heavily on 
//	// coverage voids to spot introns.
//	if (left_seg_files.size() == 0 && right_seg_files.size() == 0)
//	{
		fprintf(stderr, "Looking for junctions by island end pairings\n");
		capture_island_ends(it,
							rt,
							all_seg_files, 
							cov_juncs,
							min_intron_length,
							max_intron_length);
		fprintf(stderr, "done\n");
//	}
	
	juncs.insert(cov_juncs.begin(), cov_juncs.end());
	juncs.insert(seg_juncs.begin(), seg_juncs.end());
	
	
	fprintf(stderr, "Reporting potential splice junctions...");
	for(std::set<Junction>::iterator itr = juncs.begin();
		itr != juncs.end();
		++itr)
	{
		const char* ref_name = rt.get_name(itr->refid);
		
		fprintf(juncs_out,
				"%s\t%d\t%d\t%c\n",
				ref_name,
				itr->left,
				itr->right,
				itr->antisense ? '-' : '+');
	}
	fprintf (stderr, "done\n");
	fprintf (stderr, "Reported %d total possible splices\n", (int)juncs.size());
}

int main(int argc, char** argv)
{
    int parse_ret = parse_options(argc,argv);
    if (parse_ret)
        return parse_ret;
    
	if(optind >= argc)
    {
        print_usage();
        return 1;
    }
	
	string ref_file_name = argv[optind++];
	
	if(optind >= argc)
    {
        print_usage();
        return 1;
    }
    
    string juncs_file_name = argv[optind++];
	
    if(optind >= argc)
    {
        print_usage();
        return 1;
    }
    
    string left_segment_file_list = argv[optind++];

    
	string right_segment_file_list; 
	if (optind < argc)
		right_segment_file_list = argv[optind++];
	
    // Open the approppriate files
    
	ifstream ref_stream(ref_file_name.c_str(), ifstream::in);
    if (!ref_stream.good())
    {
        fprintf(stderr, "Error: cannot open %s for reading\n",
                ref_file_name.c_str());
        exit(1);
    }
    
	
	FILE* juncs_file = fopen(juncs_file_name.c_str(), "w");
	if (!juncs_file)
    {
        fprintf(stderr, "Error: cannot open %s for writing\n",
                juncs_file_name.c_str());
        exit(1);
    }

    
    vector<string> left_segment_file_names;
    vector<FILE*> left_segment_files;
    tokenize(left_segment_file_list, ",",left_segment_file_names);
    for (size_t i = 0; i < left_segment_file_names.size(); ++i)
    {
        FILE* seg_file = fopen(left_segment_file_names[i].c_str(), "r");
        if (seg_file == NULL)
        {
            fprintf(stderr, "Error: cannot open segment map file %s for reading\n",
                    left_segment_file_names[i].c_str());
            exit(1);
        }
        left_segment_files.push_back(seg_file);
    }
    
	vector<FILE*> right_segment_files;
	
	if (right_segment_file_list != "")
	{
		vector<string> right_segment_file_names;
		
		tokenize(right_segment_file_list, ",",right_segment_file_names);
		for (size_t i = 0; i < right_segment_file_names.size(); ++i)
		{
			FILE* seg_file = fopen(right_segment_file_names[i].c_str(), "r");
			if (seg_file == NULL)
			{
				fprintf(stderr, "Error: cannot open segment map %s for reading\n",
						right_segment_file_names[i].c_str());
				exit(1);
			}
			right_segment_files.push_back(seg_file);
		}
	}
	
    driver(ref_stream, 
		   juncs_file,   
		   left_segment_files,  
		   right_segment_files);
    
    return 0;
}
